#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

#
# Note pickling is used for message serialization.
# This is just for the prototyping & fast PoC, pickling wont be used in the production.
# Instead, protobuf messages will be defined and parsed to avoid malicious pickling.
#

import argparse
import asyncio
import binascii
import datetime
import json
import logging
import os
import re
import sys
import threading
import time

from monero_glue.agent import agent_lite, agent_misc
from monero_glue.agent.agent_lite import SignAuxData
from monero_glue.messages import DebugMoneroDiagRequest, GetEntropy, Entropy
from monero_glue.xmr import common, crypto, monero, wallet, wallet_rpc, daemon_rpc
from monero_glue.xmr.enc import chacha_poly, chacha
from monero_glue.xmr.sub import addr as xmr_addr
from monero_poc.utils import misc, trace_logger
from monero_poc.utils import cli
from monero_poc.utils.misc import TrezorAddressMismatchError
from monero_poc.utils.trezor_server_proxy import TokenProxy
from monero_serialize import xmrtypes, xmrserialize, xmrboost

import coloredlogs
from cmd2 import Cmd

logger = logging.getLogger(__name__)
coloredlogs.CHROOT_FILES = []
coloredlogs.install(level=logging.WARNING, use_chroot=False)


class HostAgent(cli.BaseCli):
    """
    Host agent wrapper
    """

    prompt = "$> "

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.args = None

        self.network_type = None
        self.address = None
        self.address_info = None  # type: xmr_addr.AddrInfo
        self.address_base_info = None  # type: xmr_addr.AddrInfo
        self.priv_view = None
        self.pub_view = None
        self.pub_spend = None
        self.network_type = None
        self.account_idx = 0  # major sub-address index
        self.wallet_salt = None
        self.wallet_password = b""
        self.wallet_file = None
        self.monero_bin = None
        self.rpc_addr = "127.0.0.1:18081"
        self.rpc_passwd = None
        self.rpc_bind_port = 48084
        self.rpc_running = False
        self.rpc_ready = False

        self.trace_logger = trace_logger.Tracelogger(logger)
        self.loop = asyncio.get_event_loop()
        self.worker_loop = asyncio.new_event_loop()
        self.worker_thread = threading.Thread(
            target=self.looper, args=(self.worker_loop,)
        )
        self.worker_thread.setDaemon(True)
        self.worker_thread.start()
        self.wallet_thread = None
        self.terminating = False

        self.trezor_proxy = None  # type: TokenProxy
        self.agent = None  # type: agent_lite.Agent
        self.token_debug = False
        self.token_path = None
        self.fresh_wallet = False

        self.wallet_proxy = wallet_rpc.WalletRpc(
            self, self.rpc_bind_port, self.rpc_passwd
        )
        self.daemon_rpc = daemon_rpc.DaemonRpc()
        self.wallet_obj = wallet.Wallet(self.daemon_rpc)

    def looper(self, loop):
        """
        Main looper
        :param loop:
        :return:
        """
        asyncio.set_event_loop(loop)
        loop.run_forever()

    def submit_coro(self, coro):
        """
        Submits corroutine to the worker loop
        :param fnc:
        :param args:
        :param kwargs:
        :return:
        """
        return asyncio.run_coroutine_threadsafe(coro, self.worker_loop)

    def wait_coro(self, coro):
        """
        Runs corouting, waits for result
        :param fnc:
        :param args:
        :param kwargs:
        :return:
        """
        future = self.submit_coro(coro)
        return future.result()

    #
    # CLI
    #

    def update_intro(self):
        """
        Updates intro text for CLI header - adds version to it.
        :return:
        """
        self.intro = (
            "-" * self.get_term_width()
            + "\n    Monero Trezor agent\n"
            + "\n"
            + "-" * self.get_term_width()
        )

    def update_prompt(self):
        """
        Prompt update
        :return:
        """
        flags = []
        if not self.rpc_running:
            flags.append("R!")
        if not self.rpc_ready:
            flags.append('Loading' if not self.fresh_wallet else 'Syncing')

        flags_str = "|".join(flags)
        flags_suffix = "|" + flags_str if len(flags_str) > 0 else ""

        self.prompt = "[wallet %s%s]: " % (
            self.address[:6].decode("ascii"),
            flags_suffix,
        )

    def token_cmd(self, coro):
        try:
            res = self.wait_coro(coro)
            return res

        except Exception as e:
            print("Trezor error (e: %s)" % e)
            self.trace_logger.log(e)
            raise e

    #
    # Handlers
    #

    def check_rpc(self):
        try_debug = None
        if not self.args.debug_rpc:
            try_debug = 'Consider running with --debug-rpc flag to show RPC wallet ' \
                        'output for closer inspection and diagnosis'

        if not self.rpc_running:
            self.perror('Monero RPC wallet is not running.')
            if try_debug:
                self.poutput(try_debug)
            return False

        elif not self.rpc_ready:
            self.perror('Monero RPC wallet is not yet ready, please wait a moment')
            self.poutput('RPC wallet is not available during the blockchain scanning, it may take a while')
            if try_debug:
                self.poutput(try_debug)
            return False
        return True

    def check_address(self):
        pres = self.token_cmd(self.agent.get_address(account=self.account_idx))
        if self.address != pres.address:
            self.perror("Connected TREZOR address does not match wallet address")
            raise TrezorAddressMismatchError()

    def do_quit(self, line):
        self.terminating = True
        return super().do_quit(line)

    do_q = do_quit
    do_Q = do_quit

    # def do_account(self, line):
    #     self.set_account(int(line))

    def do_address(self, line):
        print(self.address.decode("ascii"))

    def do_ping(self, line):
        pres = self.token_cmd(self.trezor_proxy.ping())
        if pres:
            print("OK %s" % pres)

    def do_check_trezor(self, line):
        self.check_address()
        self.poutput('OK')

    def do_state(self, line):
        if not self.rpc_running:
            self.perror('RPC wallet is not running')
        else:
            self.poutput('RPC wallet: running')

        if self.rpc_ready:
            self.poutput('RPC wallet: ready')
        elif self.fresh_wallet:
            self.poutput('RPC wallet: synchronising')
        else:
            self.poutput('RPC wallet: starting')

    def do_get_watch_only(self, line):
        pres = self.token_cmd(self.agent.get_watch_only())
        print("View key:  %s" % binascii.hexlify(pres.watch_key).decode("utf8"))
        print("Address:   %s" % pres.address.decode("utf8"))

    def do_get_address(self, line):
        pres = self.token_cmd(self.agent.get_address(account=self.account_idx))
        print("Address:   %s" % pres.address.decode("utf8"))

    def do_balance(self, line):
        if not self.check_rpc():
            return

        res = self.wallet_proxy.balance()
        print("Balance: %.5f" % wallet.conv_disp_amount(res["result"]["balance"]))
        print(
            "Unlocked Balance: %.5f"
            % wallet.conv_disp_amount(res["result"]["unlocked_balance"])
        )

    def do_height(self, line):
        if not self.check_rpc():
            return

        res = self.wallet_proxy.height()
        print("Height: %s" % res["result"]["height"])

    def do_get_transfers(self, line):
        if not self.check_rpc():
            return

        res = self.wallet_proxy.get_transfers({"pool": True, "in": True, "out": True})
        print(json.dumps(res, indent=2))

    def do_rescan_bc(self, line):
        if not self.check_rpc():
            return

        res = self.wallet_proxy.rescan_bc()
        print(json.dumps(res, indent=2))

    def do_key_image_sync(self, line):
        if not self.check_rpc():
            return

        self.check_address()
        self.wait_coro(self.key_image_sync(line))

    def do_refresh(self, line):
        if not self.check_rpc():
            return

        res = self.wallet_proxy.refresh()
        print(json.dumps(res, indent=2))

    def do_transfer(self, line):
        if not self.check_rpc():
            return

        self.check_address()
        if len(line) == 0:
            print(
                "Usage: transfer [<priority>] [<ring_size>] <address> <amount> [<payment_id>]"
            )
        parts = [x for x in line.split(" ") if len(x.strip()) > 0]

        res = misc.parse_transfer_cmd(parts)
        return self.transfer_cmd(res)

    def do_sweep_dust(self, line):
        if not self.check_rpc():
            return

        res = self.wallet_proxy.sweep_dust({'do_not_relay': True})
        if "result" not in res:
            logger.error("Sweep dust error: %s" % res)
            raise ValueError("Could not transfer")

        result = res["result"]
        unsigned = binascii.unhexlify(result["unsigned_txset"])
        self.wait_coro(self.sign_unsigned(unsigned))

    def do_sweep_all(self, line):
        if not self.check_rpc():
            return

        if len(line) == 0:
            self.poutput('sweep_all [index=<N1>[,<N2>,...]] [<priority>] [<ring_size>] <address> [<payment_id>]')
            return

        params = misc.parse_sweep_all(line)
        self.sweep_cmd(params, is_all=True)

    def do_sweep_below(self, line):
        if not self.check_rpc():
            return

        if len(line) == 0:
            self.poutput('sweep_below <amount_threshold> [index=<N1>[,<N2>,...]] [<priority>] [<ring_size>] <address> [<payment_id>]')
            return

        params = misc.parse_sweep_below(line)
        self.sweep_cmd(params, is_below=True)

    def do_sweep_single(self, line):
        if not self.check_rpc():
            return

        if len(line) == 0:
            self.poutput('sweep_single [<priority>] [<ring_size>] <key_image> <address> [<payment_id>]')
            return

        params = misc.parse_sweep_below(line)
        self.sweep_cmd(params, is_single=True)

    def do_sign(self, line):
        if not self.check_rpc():
            return

        self.check_address()
        self.wait_coro(self.sign_wrap(line))

    def do_init(self, line):
        mnemonic12 = "alcohol woman abuse must during monitor noble actual mixed trade anger aisle"
        mnemonic24 = "permit universe parent weapon amused modify essay borrow tobacco budget walnut lunch consider gallery ride amazing frog forget treat market chapter velvet useless topple"

        mnemonic = mnemonic12 if len(line) == 0 or int(line) == 0 else mnemonic24
        self.trezor_proxy.client.wipe_device()
        self.trezor_proxy.client.load_device_by_mnemonic(
            mnemonic=mnemonic,
            pin="",
            passphrase_protection=False,
            label="ph4test",
            language="english",
        )

    def do_get_entropy(self, line):
        parts = line.split(' ')
        size = int(parts[0])
        path = parts[1]
        self.wait_coro(self.get_entropy(size, path))

    def do_tdeb(self, line):
        is_deb = bool(int(line))
        print("Token debug set to: %s" % is_deb)
        self.token_debug = is_deb

    def do_switch(self, line):
        path = "bridge:web01"
        if line == "udp":
            path = "udp:127.0.0.1:21324"
        elif ':' in line:
            path = line

        if 'bridge' in path:
            self.token_debug = False

        print("Switching to device: %s" % path)
        self.token_path = path
        self.wait_coro(self.connect(path))

    def do_reconnect(self, line):
        self.wait_coro(self.connect(self.token_path))

    def do_enumerate(self, line):
        from monero_glue.trezor import manager as tmanager
        r = tmanager.Trezor.enumerate()
        for x in r:
            print(x)

    def do_diag(self, line):
        m = re.match(r'^(\d+)(?:(\s+\d+)(\s+\d+)?)?', line.strip())
        if m is None:
            print('Usage: diag INS [p1 [p2]]')
            return

        diag_code = int(m.group(1))
        p1 = int(m.group(2)) if m.group(2) else None
        p2 = int(m.group(3)) if m.group(3) else None
        print("Diagnosis: %d p1: %s p2: %s" % (diag_code, p1, p2))
        msg = DebugMoneroDiagRequest(ins=diag_code, p1=p1, p2=p2)
        try:
            resp = self.wait_coro(self.trezor_proxy.call(msg))
            print(resp)

        except Exception as e:
            self.trace_logger.log(e)
            logger.warning(e)

    complete_sign = Cmd.path_complete

    #
    # Logic
    #

    def set_network_type(self, ntype):
        self.network_type = ntype
        self.agent.network_type = ntype

    async def is_connected(self):
        """
        Returns True if Trezor is connected
        :return:
        """
        try:
            await self.trezor_proxy.ping()
            return True

        except Exception as e:
            return False

    async def load_watchonly(self):
        """
        Loads watch-only credentials from connected Trezor
        :return:
        """
        if not await self.is_connected():
            logger.error("Trezor is not connected")
            raise agent_misc.TrezorNotRunning("Could not load watch-only credentials")

        try:
            print(
                "Loading watch-only credentials from Trezor. Please, confirm the request on Trezor."
            )
            self.set_network_type(
                monero.NetworkTypes.TESTNET
                if self.args.testnet
                else monero.NetworkTypes.MAINNET
            )

            res = await self.agent.get_watch_only()  # type: messages.MoneroWatchKey

            self.priv_view = crypto.decodeint(res.watch_key)
            self.set_address(res.address)
            await self.open_with_keys(self.priv_view, self.address)

        except Exception as e:
            raise ValueError(e)

    async def get_entropy(self, size, path):
        """
        Loads entropy from the device, writes to the path
        :param size:
        :param path:
        :return:
        """
        logger.info('Loading entropy of %s B to %s' % (size, path))
        with open(path, 'wb+') as fh:
            csize = 0
            while csize < size:
                req = GetEntropy(size=1024*10)
                res = await self.trezor_proxy.call_in_session(req)  # type: Entropy
                csize += len(res.entropy)
                logger.debug(' .. loaded %s / %s (%s %%)' % (csize, size, 100.0*csize / size))
                fh.write(res.entropy)
        logger.info('Entropy loading done')

    def load_params(self):
        """
        Args
        :return:
        """
        if self.args.rpc_addr:
            self.rpc_addr = self.args.rpc_addr
        if self.args.watch_wallet:
            self.wallet_file = self.args.watch_wallet
        if self.args.monero_bin:
            self.monero_bin = self.args.monero_bin

    async def choose_trezor(self):
        from monero_glue.trezor import manager as tmanager
        r = tmanager.Trezor.enumerate()
        noptions = len(r)

        if noptions == 0:
            self.perror("No TREZOR device found")
            raise EnvironmentError("No usable device")

        elif noptions == 1:
            self.poutput("Detected TREZOR: %s" % r[0])
            return str(r[0])

        choices = [(i, str(r[i])) for i in range(noptions)]
        res = self.select(choices, "Please select from connected devices: ")
        return str(r[res])

    def monkey_patch_trezorlib(self):
        logger.info('Monkey-patching trezorlib with the current messages versions fromt trezor-common')
        try:
            import trezorlib
            from monero_glue import protobuf
            from monero_glue import messages
            from trezorlib import protobuf as tprotobuf

            tprotobuf.UVarintType = protobuf.UVarintType
            tprotobuf.SVarintType = protobuf.SVarintType
            tprotobuf.BoolType = protobuf.BoolType
            tprotobuf.BytesType = protobuf.BytesType
            tprotobuf.UnicodeType = protobuf.UnicodeType
            tprotobuf.MessageType = protobuf.MessageType
            trezorlib.messages = messages

            from trezorlib import mapping
            trezorlib.mapping.map_type_to_class = {}
            trezorlib.mapping.map_class_to_type = {}
            trezorlib.mapping.build_map()

        except Exception as e:
            logger.error('Monkey patching error: %s' % e)

    async def connect(self, path=None):
        """
        Connects to the trezor
        :return:
        """
        if not self.args.poc or (self.args.trezor_path or path):
            if self.args.patch_client:
                self.monkey_patch_trezorlib()

            from monero_glue.trezor import manager as tmanager

            t_path = path if path else self.args.trezor_path
            if t_path is None or len(t_path) == 0:
                self.token_path = await self.choose_trezor()
                t_path = self.token_path

            self.trezor_proxy = tmanager.Trezor(
                path=t_path, debug=self.token_debug
            )

        else:
            self.trezor_proxy = TokenProxy()

        ntype = self.agent.network_type if self.agent else self.network_type
        self.agent = agent_lite.Agent(self.trezor_proxy, network_type=ntype, slip0010=self.args.slip0010)

    async def open_account(self):
        """
        Opens the watch only account
        :return:
        """
        creds_passed = self.args.view_key is not None and self.args.address is not None
        account_file_set = self.args.account_file is not None
        account_file_ex = (
            os.path.exists(self.args.account_file) if account_file_set else False
        )

        if creds_passed:
            await self.open_account_passed()
        elif account_file_ex:
            await self.open_account_file(self.args.account_file)
        else:
            self.load_params()
            await self.check_params(True)
            await self.load_watchonly()

        self.load_params()
        if account_file_set and not account_file_ex:
            await self.check_params(True)
            self.wallet_password = await self.prompt_password(True)

        # Create watch only wallet file for monero-wallet-rpc
        await self.ensure_watch_only()

        # Write acquired data to the account file
        if account_file_set and not account_file_ex:
            await self.save_account(self.args.account_file)

        print(
            "Public spend key: %s"
            % binascii.hexlify(crypto.encodepoint(self.pub_spend)).decode("ascii")
        )
        print(
            "Public view key : %s"
            % binascii.hexlify(crypto.encodepoint(self.pub_view)).decode("ascii")
        )
        print("Address:          %s" % self.address.decode("utf8"))
        self.update_intro()
        self.update_prompt()

    async def check_params(self, new_wallet=False):
        """
        All params correctly entered?
        :return:
        """
        if not new_wallet:
            return
        if self.args.sign is not None:
            return
        if self.wallet_file is None:
            logger.error(
                "--watch-wallet file is not set. Please specify path where to create the monero watch wallet"
            )
            sys.exit(1)
        if self.monero_bin is None:
            logger.error(
                "--monero-bin is not set. Please specify path to the monero binaries"
            )
            sys.exit(1)

    async def prompt_password(self, new_wallet=False):
        """
        Prompts password for a new wallet
        :param new_wallet:
        :return:
        """
        if new_wallet:
            passwd = self.ask_password(
                "Creating a new wallet. Please, enter the password: ", True
            )
        else:
            passwd = self.ask_password("Please, enter the wallet password: ", False)
        return passwd.encode("utf8")

    async def save_account(self, file):
        """
        Stores account data
        :param file:
        :return:
        """
        if self.wallet_salt is None:
            self.wallet_salt = crypto.random_bytes(32)

        # Wallet view key encryption
        wallet_enc_key = misc.wallet_enc_key(self.wallet_salt, self.wallet_password)
        ciphertext = chacha_poly.encrypt_pack(wallet_enc_key, crypto.encodeint(self.priv_view))

        with open(file, "w") as fh:
            data = {
                "view_key_enc": binascii.hexlify(ciphertext).decode("ascii"),
                "address": self.address.decode("ascii"),
                "network_type": self.network_type,
                "wallet_salt": binascii.hexlify(self.wallet_salt).decode("ascii"),
                "rpc_addr": self.rpc_addr,
                "wallet_file": self.wallet_file,
                "monero_bin": self.monero_bin,
            }
            json.dump(data, fh, indent=2)

    async def check_existing_wallet_file(self, key_file):
        """
        Checks existing wallet file correctness
        :param key_file:
        :return:
        """
        wl = await wallet.load_keys_file(key_file, self.wallet_password)
        addr = wl["key_data"]["m_keys"]["m_account_address"]
        spend_pub = addr["m_spend_public_key"]
        view_pub = addr["m_view_public_key"]

        match = spend_pub == crypto.encodepoint(
            self.pub_spend
        ) and view_pub == crypto.encodepoint(self.pub_view)
        net_ver = monero.net_version(self.network_type, False)
        addr = monero.encode_addr(net_ver, spend_pub, view_pub)
        return addr, match

    async def wallet_restore_param(self):
        self.poutput("Creating a new wallet file, please enter the blockchain height to start a restore of the wallet.")
        self.poutput("  - Restore height should be a little less than your first incoming transaction to the wallet.")
        self.poutput("  - If the wallet was never used before, enter \"-\"")
        self.poutput("  - If you are not sure enter \"0\" to start from the beginning (may take few minutes)")
        self.poutput("  - You may enter also a date in the format YYYY-MM-DD\n")

        height = 0
        sure_stage = False

        while True:
            if sure_stage:
                if self.ask_proceed_quit("The height: %s. Is it correct? (y/n) " % height) == self.PROCEED_YES:
                    break

            height = misc.py_raw_input("Restore height: ").strip().lower()
            if height == '-':
                height = await self.wallet_obj.get_height()
                sure_stage = True
                continue

            elif len(height) == 0:
                height = 0
                sure_stage = True
                continue

            m = re.match(r'^(\d{4})-(\d{1,2})-(\d{1,2})$', height)
            if m:
                year, month, day = int(m.group(1)), int(m.group(2)), int(m.group(3))
                try:
                    height = await self.wallet_obj.get_blockchain_height_by_date(year, month, day)
                    sure_stage = True
                    continue

                except Exception as e:
                    logger.warning('Could not resolve date to the height: %s' % e)
                    self.trace_logger.log(e)
                    r = self.ask_proceed_quit("Could not resolve date to height. Do you want to try again? (y/n) ")
                    if r == self.PROCEED_YES:
                        continue
                    else:
                        return 0

            else:
                try:
                    height = int(height)
                    sure_stage = True
                    continue

                except Exception as e:
                    self.poutput("Invalid format")
                    sure_stage = False
                    continue
        return height

    async def ensure_watch_only(self):
        """
        Ensures watch only wallet for monero exists
        :return:
        """
        if self.wallet_file is None:
            return

        key_file = "%s.keys" % self.wallet_file
        if os.path.exists(key_file):
            logger.debug("Watch only wallet key file exists: %s" % key_file)
            match, addr = False, None
            try:
                addr, match = await self.check_existing_wallet_file(key_file)
            except Exception as e:
                logger.error("Wallet key file processing exception: %s" % e)

            if not match:
                logger.error("Key file address is not correct: %s" % addr)
                print("Please, move the file so Agent can create correct key file")
                sys.exit(2)
            return

        self.fresh_wallet = True
        account_keys = xmrtypes.AccountKeys()
        key_data = wallet.WalletKeyData()

        restore_height = await self.wallet_restore_param()
        self.poutput('Wallet restore height: %s' % restore_height)

        wallet_data = wallet.WalletKeyFile()
        wallet_data.key_data = key_data
        wallet_data.watch_only = 1
        wallet_data.testnet = self.network_type == monero.NetworkTypes.TESTNET
        wallet_data.refresh_height = restore_height

        key_data.m_creation_timestamp = int(time.time())
        key_data.m_keys = account_keys

        account_keys.m_account_address = xmrtypes.AccountPublicAddress(
            m_spend_public_key=crypto.encodepoint(self.pub_spend),
            m_view_public_key=crypto.encodepoint(self.pub_view),
        )
        account_keys.m_spend_secret_key = crypto.encodeint(crypto.sc_0())
        account_keys.m_view_secret_key = crypto.encodeint(self.priv_view)

        await wallet.save_keys_file(key_file, self.wallet_password, wallet_data)
        logger.debug("Watch-only wallet keys generated: %s" % key_file)

    async def open_account_passed(self):
        """
        Loads passed credentials
        :return:
        """
        priv_view = self.args.view_key.encode("ascii")
        self.priv_view = crypto.b16_to_scalar(priv_view)
        self.set_network_type(
            monero.NetworkTypes.TESTNET
            if self.args.testnet
            else monero.NetworkTypes.MAINNET
        )
        self.set_address(self.args.address.encode("ascii"))
        self.wallet_file = self.args.watch_wallet
        self.monero_bin = self.args.monero_bin
        await self.open_with_keys(self.priv_view, self.address)

    async def open_account_file(self, file):
        """
        Opens account file
        :param file:
        :return:
        """
        with open(file) as fh:
            js = json.load(fh)

        # Wallet key encryption
        self.wallet_password = await self.prompt_password()
        self.wallet_salt = common.defvalkey(js, "wallet_salt")
        if self.wallet_salt is None:
            self.wallet_salt = crypto.random_bytes(32)
        else:
            self.wallet_salt = binascii.unhexlify(self.wallet_salt)

        # Wallet view key dec.
        if "view_key" in js:
            self.priv_view = crypto.b16_to_scalar(js["view_key"].encode("utf8"))

        elif "view_key_enc" in js:
            wallet_enc_key = misc.wallet_enc_key(self.wallet_salt, self.wallet_password)
            plain = chacha_poly.decrypt_pack(wallet_enc_key, binascii.unhexlify(js["view_key_enc"]))
            self.priv_view = crypto.decodeint(plain)

        self.wallet_file = js["wallet_file"]
        self.monero_bin = js["monero_bin"]
        self.set_network_type(js["network_type"])
        self.set_address(js["address"].encode("utf8"))
        self.rpc_addr = js["rpc_addr"]

        await self.open_with_keys(self.priv_view, self.address)

    async def open_with_keys(self, view_key, address):
        """
        Processess view key private + address
        :param view_key:
        :param address:
        :return:
        """
        self.pub_view = crypto.scalarmult_base(view_key)

        addr_info = monero.decode_addr(address)
        self.pub_spend = crypto.decodepoint(addr_info.spend_key)

        if not crypto.point_eq(self.pub_view, crypto.decodepoint(addr_info.view_key)):
            raise ValueError(
                "Computed view public key does not match the one from address"
            )

    def set_address(self, address):
        self.address = address
        self.address_info = monero.decode_addr(self.address)
        self.address_base_info = monero.decode_addr(self.address)
        self.recompute_address()

    def recompute_address(self):
        D, C = monero.generate_sub_address_keys(
            self.priv_view, crypto.decodepoint(self.address_base_info.spend_key), self.account_idx, 0)

        self.address_info.recompute_sub(crypto.encodepoint(D), crypto.encodepoint(C), self.account_idx)
        self.address = self.address_info.addr
        self.update_prompt()

    def set_account(self, new_accound_idx):
        self.account_idx = new_accound_idx
        self.recompute_address()

    def wallet_rpc_main(self, *args, **kwargs):
        """
        Wallet RPC thread
        :return:
        """
        rpc_cmd = os.path.join(self.monero_bin, "monero-wallet-rpc")
        if not os.path.exists(rpc_cmd):
            logger.error("Wallet rpc binary not found: %s" % rpc_cmd)
            sys.exit(1)

        self.rpc_passwd = misc.gen_simple_passwd(16)
        self.wallet_proxy.set_creds(["trezor", self.rpc_passwd])

        args = [
            "--daemon-address %s" % misc.escape_shell(self.rpc_addr),
            "--wallet-file %s" % misc.escape_shell(self.wallet_file),
            "--prompt-for-password",
            "--rpc-login=trezor",
            "--rpc-bind-port %s" % int(self.rpc_bind_port),
        ]

        if self.args.testnet or self.network_type == monero.NetworkTypes.TESTNET:
            args.append("--testnet")
        if self.args.debug_rpc:
            logger.debug('RPC credentials: trezor:%s' % self.rpc_passwd)

        def preexec_function():
            os.setpgrp()

        cmd = "%s %s" % (rpc_cmd, " ".join(args))
        feeder = misc.Feeder()
        p = misc.run(
            cmd,
            input=feeder,
            async_=True,
            stdout=misc.Capture(timeout=0.1, buffer_size=1),
            stderr=misc.Capture(timeout=0.1, buffer_size=1),
            cwd=os.getcwd(),
            env=None,
            shell=True,
            preexec_fn=preexec_function
        )

        ret_code = 1
        out_acc, err_acc = [], []
        out_cur, err_cur = [""], [""]
        logline_pattern = re.compile(r'^([^\s]+)\s+([^\s]+)\s+\[(.+?)\]\s+([^\s]+?)\s+')
        passwd_set = False

        def log_parse(line):
            m = logline_pattern.match(line)
            if m is None:
                return None
            sev = m.group(4).lower()
            sevnum = None
            if sev == 'error':
                sevnum = logging.ERROR
            elif sev == 'warn' or sev == 'warning':
                sevnum = logging.WARNING
            elif sev == 'info':
                sevnum = logging.INFO
            elif sev == 'debug':
                sevnum = logging.DEBUG
            return m.group(1), m.group(2), m.group(3), m.group(4), sevnum

        def process_line(line, is_err=False):
            dst = err_acc if is_err else out_acc
            dst.append(line)
            line_parsed = log_parse(line)
            line_printed = False
            if line_parsed and line_parsed[4] and line_parsed[4] >= logging.ERROR:
                logger.error('RPC_%s: %s' % ('ERR' if is_err else 'OUT', line))
                line_printed = True

            if not line_printed and self.args.debug_rpc:
                logger.debug('RPC_%s: %s' % ('ERR' if is_err else 'OUT', line))

            line_low = line.lower()
            if 'starting wallet' in line_low:
                self.rpc_ready = True
                self.on_rpc_ready()

        def add_output(buffers, is_err=False):
            buffers = [x.decode("utf8") for x in buffers]
            lines = [""]

            dst_cur = err_cur if is_err else out_cur
            for x in buffers:
                clines = [v.strip('\r') for v in x.split('\n')]
                lines[-1] += clines[0]
                lines.extend(clines[1:])

            dst_cur[0] += lines[0]
            nlines = len(lines)
            if nlines > 1:
                process_line(dst_cur[0])
                dst_cur[0] = ''

            for line in lines[1:-1]:
                process_line(line, is_err)

            dst_cur[0] = lines[-1] or ''

        try:
            while len(p.commands) == 0:
                time.sleep(0.15)

            self.rpc_running = True
            self.update_prompt()

            while p.commands[0].returncode is None:
                if not passwd_set:
                    passwd_set = True
                    feeder.feed(self.wallet_password)
                    feeder.feed("\n")
                    feeder.feed(self.rpc_passwd)
                    feeder.feed("\n")

                out, err = p.stdout.read(-1, False), p.stderr.read(-1, False)
                if not common.is_empty(out):
                    add_output([out])
                if not common.is_empty(err):
                    add_output([err], True)

                p.commands[0].poll()
                if self.terminating and p.commands[0].returncode is None:
                    feeder.feed("quit\n\n")
                    misc.sarge_sigint(p.commands[0])
                    p.close()
                if not common.is_empty(out) or not common.is_empty(err):
                    continue
                time.sleep(0.01)

            ret_code = p.commands[0].returncode
            add_output([p.stdout.read(-1, False)])
            add_output([p.stderr.read(-1, False)], True)
            self.rpc_running = False
            self.update_prompt()

            if not self.terminating:
                logger.error("Wallet RPC ended prematurely with code: %s" % ret_code)
                logger.info("Command: %s" % cmd)
                logger.info("Std out: %s" % "\n".join(out_acc))
                logger.info("Error out: %s" % "\n".join(err_acc))

        except Exception as e:
            logger.error("Exception in wallet RPC command: %s" % e)
            self.trace_logger.log(e)

    def shutdown_rpc(self):
        """
        Waits for rpc shutdown
        :return:
        """
        if self.args.wallet_rpc_addr:  # using already running rpc
            return
        if not self.rpc_running:
            return

        # Gracegul stop with save
        try:
            self.wallet_proxy.stop_wallet()
            self.terminating = True
            time.sleep(1)
        except Exception as e:
            logger.warning("Stopping wallet failed: %s" % e)

        # Terminating with sigint
        logger.info("Waiting for wallet-RPC to terminate...")
        self.terminating = True
        while self.rpc_running:
            time.sleep(0.1)

    def on_rpc_ready(self):
        """
        Called when RPC is started
        :return:
        """
        self.update_prompt()

    async def wallet_rpc(self):
        """
        Starts wallet RPC server
        :return:
        """
        if self.args.wallet_rpc_addr:  # using existing RPC?
            self.daemon_rpc.set_addr(self.args.rpc_addr)
            self.wallet_proxy.set_addr(self.args.wallet_rpc_addr)
            self.wallet_proxy.set_creds(self.args.wallet_rpc_creds)
            self.rpc_running = True
            self.update_prompt()
            return

        self.wallet_thread = threading.Thread(target=self.wallet_rpc_main, args=(None,))
        self.wallet_thread.setDaemon(False)
        self.wallet_thread.start()

    async def entry(self):
        """
        Entry point
        :return:
        """
        if self.args.debug:
            coloredlogs.install(level=logging.DEBUG, use_chroot=False)
        misc.install_sarge_filter()

        await self.connect()
        await self.open_account()

        if self.args.sign:
            res = await self.sign_wrap(self.args.sign)
            return res if isinstance(res, int) else 0

        await self.wallet_rpc()

        self.update_intro()
        self.cmdloop()
        self.shutdown_rpc()
        logger.info("Terminating")

    #
    # Sign op
    #

    def handle_address_input(self, address, payment_id=None):
        try:
            address = address.encode("ascii")  # make bytes
        except:
            pass

        if payment_id and (len(payment_id) != 16 and len(payment_id) != 64):
            self.perror("Payment ID can be either 8B or 32B long")
            raise ValueError("Invalid payment ID")

        try:
            addr_info = monero.decode_addr(address)

            if addr_info.is_integrated and payment_id:
                raise ValueError('Address is integrated (contains payment id), redundant payment_id provided')

            if payment_id:
                payment_id = binascii.unhexlify(payment_id)

        except Exception as e:
            self.perror("Address invalid: %s, error: %s " % (address, e))
            raise

        return addr_info, payment_id

    def transfer_cmd(self, parts):
        """
        Transfer logic
        :param parts:
        :return:
        """
        priority, mixin, payment_id = parts.priority, parts.mixin, parts.payment_id

        aux_data = SignAuxData()
        aux_data.destinations = []
        addr_amnt = parts.address_amounts
        destinations = []
        new_payment_id = None

        for idx, cur in enumerate(addr_amnt):
            addr_info, tmp_payment_id = self.handle_address_input(cur[0], payment_id)
            amount_atomic = misc.amount_to_uint64(cur[1])
            aux_data.destinations.append((addr_info, amount_atomic))
            destinations.append({
                'amount': amount_atomic,
                'address': addr_info.addr.decode('ascii'),
            })

            if tmp_payment_id:
                new_payment_id = tmp_payment_id

            print("Sending %s monero to %s" % (cur[1], cur[0]))

        print(
            "Priority: %s, mixin: %s, payment_id: %s"
            % (
                priority if priority else "default",
                mixin if mixin else "default",
                binascii.hexlify(new_payment_id).decode('ascii') if new_payment_id else "-",
            )
        )

        ask_res = self.ask_proceed_quit("Do you confirm (y/n) ? ")
        if ask_res != self.PROCEED_YES:
            return

        params = {
            "destinations": destinations,
            "account_index": self.account_idx,
            "subaddr_indices": parts.indices,
            "unlock_time": 0,
            "get_tx_keys": True,
            "do_not_relay": True,
            "get_tx_hex": False,
            "get_tx_metadata": False,
        }

        if priority is not None:
            params["priority"] = priority

        if mixin is not None:
            params["mixin"] = mixin

        if new_payment_id is not None:
            params["payment_id"] = binascii.hexlify(new_payment_id).decode('ascii')

        # Call RPC to prepare unsigned transaction
        self.transfer_params(params, aux_data)

    def sweep_params(self, parts, is_all=False, is_below=False, is_single=False):
        params = {
            'do_not_relay': True
        }

        if parts.priority is not None:
            params["priority"] = parts.priority

        if parts.mixin is not None:
            params["ring_size"] = parts.mixin

        if parts.payment_id is not None:
            params["payment_id"] = parts.payment_id

        if parts.amount_threshold is not None:
            params["below_amount"] = misc.amount_to_uint64(parts.amount_threshold)

        if parts.key_image is not None:
            params["key_image"] = parts.key_image

        if not is_single:
            params['account_index'] = self.account_idx
            params['subaddr_indices'] = parts.indices
            params['address'] = parts.address

        return params

    def sweep_cmd(self, parts, is_all=False, is_below=False, is_single=False):
        params = self.sweep_params(parts, is_all, is_below, is_single)
        if is_single:
            res = self.wallet_proxy.sweep_single(params)
        else:
            res = self.wallet_proxy.sweep_all(params)

        if "result" not in res:
            logger.error("Sweep error: %s" % res)
            raise ValueError("Could not transfer")
        result = res["result"]

        amounts = common.defvalkey(result, 'amount_list', [common.defvalkey(result, 'amount')])
        fees = common.defvalkey(result, 'fee_list', [common.defvalkey(result, 'fee')])
        for idx in range(len(amounts)):
            st = "Amount: %s" % wallet.conv_disp_amount(amounts[idx])
            if idx < len(fees):
                st += ", Fee: %s" % wallet.conv_disp_amount(fees[idx])
            self.poutput(st)

        ask_res = self.ask_proceed_quit("Do you confirm (y/n) ? ")
        if ask_res != self.PROCEED_YES:
            return

        unsigned = binascii.unhexlify(result["unsigned_txset"])
        self.wait_coro(self.sign_unsigned(unsigned))

    def transfer_params(self, params, aux_data=None):
        res = self.wallet_proxy.transfer_split(params)
        if "result" not in res:
            logger.error("Transfer error: %s" % res)
            raise ValueError("Could not transfer")

        result = res["result"]

        amounts = common.defvalkey(result, 'amount_list', [common.defvalkey(result, 'amount')])
        fees = common.defvalkey(result, 'fee_list', [common.defvalkey(result, 'fee')])
        for idx in range(len(amounts)):
            st = "Amount: %s" % wallet.conv_disp_amount(amounts[idx])
            if idx < len(fees):
                st += ", Fee: %s" % wallet.conv_disp_amount(fees[idx])
            self.poutput(st)

        ask_res = self.ask_proceed_quit("Do you confirm (y/n) ? ")
        if ask_res != self.PROCEED_YES:
            return

        if "unsigned_txset" not in result:
            logger.error(
                "Unsigned transaction not found in the response. "
                "Please make sure you are using compatible monero-wallet-rpc"
            )
            logger.debug(res)
            return

        unsigned = binascii.unhexlify(result["unsigned_txset"])
        self.wait_coro(self.sign_unsigned(unsigned, aux_data))

    async def sign_unsigned(self, unsigned_txset, aux_data=None):
        """
        Signs unsigned txset with the Trezor
        :param unsigned_txset:
        :return:
        """
        res = await self.sign_wrap(fdata=unsigned_txset, aux_data=aux_data)
        if isinstance(res, int):
            logger.error("Error")
            return

        print("Transaction has been signed. ")
        ask_res = self.ask_proceed_quit("Do you wish to submit (y/n) ? ")
        if ask_res != self.PROCEED_YES:
            return

        params = {"tx_data_hex": binascii.hexlify(res).decode("ascii")}

        res = self.wallet_proxy.submit_transfer(params)
        try:
            if len(res["result"]["tx_hash_list"]) == 0:
                raise ValueError("Transaction submit failed")

            print("SUCCESS: Transaction has been submitted!")

        except Exception as e:
            logger.debug("Res: %s" % res)
            print("Transaction submit failed: %s" % e)

    async def sign_wrap(self, file=None, fdata=None, aux_data=None):
        """
        Sign wrapper
        :param file:
        :param fdata:
        :param aux_data:
        :return:
        """
        if not self.priv_view:
            logger.error("View key not set, cannot sign")
            return -3

        try:
            return await self.sign(file, fdata, aux_data=aux_data)

        except agent_misc.TrezorReturnedError as e:
            self.trace_logger.log(e)
            print("Trezor returned an error: %s" % e)
            return 1

        except agent_misc.TrezorNotRunning as e:
            logger.error("Trezor server is not running")
            return 2

    async def sign(self, file=None, fdata=None, aux_data=None):
        """
        Performs TX signature
        :param file:
        :param fdata:
        :param aux_data:
        :return:
        """
        try:
            await self.trezor_proxy.ping()
        except Exception as e:
            raise agent_misc.TrezorNotRunning(e)

        if file and not os.path.exists(file):
            raise ValueError("Could not find unsigned transaction file")

        data = fdata
        if data is None:
            with open(file, "rb") as fh:
                data = fh.read()

        msg = await wallet.load_unsigned_tx(self.priv_view, data)

        # Key image sync
        # key_images = await self.agent.import_outputs(msg.transfers)
        # For now sync only spent key images to the hot wallet.
        key_images = [td.m_key_image for td in msg.transfers]
        max_ki_size = 0
        if len(key_images) == 0:
            logger.info('Wallet did not return transfer list :/')
            for tx in msg.txes:
                for idx in range(len(tx.selected_transfers)):
                    max_ki_size = max(max_ki_size, tx.selected_transfers[idx])
            key_images = [crypto.identity(True)] * (max_ki_size + 1)

        txes = []
        pendings = []
        for tx in msg.txes:  # type: xmrtypes.TxConstructionData
            print("Signing transaction with Trezor")
            print("Please check the Trezor and confirm / reject the transaction\n")

            res = await self.agent.sign_transaction_data(tx, aux_data=aux_data)
            cdata = self.agent.last_transaction_data()
            await self.store_cdata(cdata, res, tx, msg.transfers)

            # obj = await xmrobj.dump_message(None, res)
            # print(xmrjson.json_dumps(obj, indent=2))

            # Key image sync for spent TXOs
            # Updating only spent.
            for idx in range(len(tx.selected_transfers)):
                idx_mapped = cdata.source_permutation[idx]
                key_images[tx.selected_transfers[idx_mapped]] = res.vin[idx].k_image

            txes.append(await self.agent.serialize_tx(res))
            pending = wallet.construct_pending_tsx(res, tx)
            pendings.append(pending)

        # Key images array has to cover all transfers sent.
        # Watch only wallet does not have key images.
        signed_tx = xmrtypes.SignedTxSet(ptx=pendings, key_images=key_images)
        signed_data = await wallet.dump_signed_tx(self.priv_view, signed_tx)
        with open("signed_monero_tx", "wb+") as fh:
            fh.write(signed_data)
            print("Signed transaction file: signed_monero_tx")

        print(
            "Key images: %s"
            % [binascii.hexlify(ff).decode("utf8") for ff in key_images]
        )
        for idx, tx in enumerate(txes):
            fname = "transaction_%02d" % idx
            with open(fname, "wb+") as fh:
                fh.write(tx)

            # relay_fname = 'transaction_%02d_relay.sh' % idx
            # hex_ctx = binascii.hexlify(tx).decode('utf8')
            # with open(relay_fname, 'w+') as fh:
            #     fh.write('#!/bin/bash\n')
            #     fh.write('curl -X POST http://%s/sendrawtransaction '
            #              '-d \'{"tx_as_hex":"%s", "do_not_relay":false}\' '
            #              '-H \'Content-Type: application/json\'\n' % (self.args.rpc_addr, hex_ctx))
            #
            # print('Transaction %02d stored to %s, relay script: %s' % (idx, fname, relay_fname))

            # Relay:
            # payload = {'tx_as_hex': hex_ctx, 'do_not_relay': False}
            # resp = requests.post('http://%s/sendrawtransaction' % (self.args.rpc_addr, ), json=payload)
            # print('Relay response: %s' % resp.json())

        # print('Please note that by manual relaying hot wallet key images get out of sync')
        return signed_data

    async def store_cdata(self, cdata, signed_tx, tx, transfers):
        """
        Stores transaction data for later usage.
            - cdata.enc_salt1, cdata.enc_salt2, cdata.enc_keys
            - tx_keys are AEAD protected, key derived from spend key - only token can open.
            - construction data for further proofs.

        :param cdata:
        :param signed_tx:
        :param tx:
        :param transfers:
        :return:
        """
        hash = cdata.tx_prefix_hash
        prefix = binascii.hexlify(hash[:12])

        tx_key_salt = crypto.random_bytes(32)
        tx_key_inp = hash + crypto.encodeint(self.priv_view)
        tx_view_key = crypto.pbkdf2(tx_key_inp, tx_key_salt, 2048)

        unsigned_data = xmrtypes.UnsignedTxSet()
        unsigned_data.txes = [tx]
        unsigned_data.transfers = transfers if transfers is not None else []

        writer = xmrserialize.MemoryReaderWriter()
        ar = xmrboost.Archive(writer, True)
        await ar.root()
        await ar.message(unsigned_data)

        unsigned_key = crypto.keccak_2hash(b'unsigned;' + tx_view_key)
        ciphertext = chacha_poly.encrypt_pack(unsigned_key, bytes(writer.get_buffer()))

        # Serialize signed transaction
        writer = xmrserialize.MemoryReaderWriter()
        ar = xmrserialize.Archive(writer, True)
        await ar.root()
        await ar.message(signed_tx)
        signed_tx_bytes = writer.get_buffer()
        signed_tx_hmac_key = crypto.keccak_2hash(b'hmac;' + tx_view_key)
        signed_tx_hmac = crypto.compute_hmac(signed_tx_hmac_key, signed_tx_bytes)

        try:
            js = {
                "time": int(time.time()),
                "hash": binascii.hexlify(hash).decode("ascii"),
                "enc_salt1": binascii.hexlify(cdata.enc_salt1).decode("ascii"),
                "enc_salt2": binascii.hexlify(cdata.enc_salt2).decode("ascii"),
                "tx_keys": binascii.hexlify(cdata.enc_keys).decode("ascii"),
                "unsigned_data": binascii.hexlify(ciphertext).decode("ascii"),
                "tx_salt": binascii.hexlify(tx_key_salt).decode("ascii"),
                "tx_signed": binascii.hexlify(signed_tx_bytes).decode("ascii"),
                "tx_signed_hmac": binascii.hexlify(signed_tx_hmac).decode("ascii"),
            }

            with open("transaction_%s.json" % prefix.decode("ascii"), "w") as fh:
                json.dump(js, fh, indent=2)
                fh.write("\n")

        except Exception as e:
            self.trace_logger.log(e)
            print(
                "Unable to save transaction data for transaction %s"
                % binascii.hexlify(hash).decode("ascii")
            )

    async def key_image_sync(self, line):
        """
        Key image sync with Trezor
        :param line:
        :return:
        """
        res = self.wallet_proxy.export_outputs()
        outputs_data_hex = res["result"]["outputs_data_hex"]

        outs_data = binascii.unhexlify(outputs_data_hex)
        exps = await wallet.load_exported_outputs(self.priv_view, outs_data)

        # Check if for this address
        match = exps.m_spend_public_key == crypto.encodepoint(
            self.pub_spend
        ) and exps.m_view_public_key == crypto.encodepoint(self.pub_view)
        net_ver = monero.net_version(self.network_type, False)
        addr = monero.encode_addr(
            net_ver, exps.m_spend_public_key, exps.m_view_public_key
        )
        if not match:
            logger.error(
                "Exported outputs from different wallet: %s" % addr.decode("ascii")
            )
            return

        self.poutput("Exported outputs loaded.")
        self.poutput("Please confirm the key image sync on the Trezor ")
        res = await self.agent.import_outputs(exps.tds)

        # Generate import key image requests
        key_images = []
        for kie in res:
            key_images.append(
                {
                    "key_image": binascii.hexlify(kie[0]).decode("ascii"),
                    "signature": binascii.hexlify(kie[1][0] + kie[1][1]).decode(
                        "ascii"
                    ),
                }
            )

        import_req = {"signed_key_images": key_images}

        res = self.wallet_proxy.import_key_images(import_req)
        print("Height: %s" % res["result"]["height"])
        print("Spent: %.5f" % wallet.conv_disp_amount(res["result"]["spent"]))
        print("Unspent: %.5f" % wallet.conv_disp_amount(res["result"]["unspent"]))

    async def main(self):
        """
        Entry point
        :return:
        """
        parser = argparse.ArgumentParser(description="Trezor Agent")

        parser.add_argument("--address", dest="address", help="Full address")

        parser.add_argument(
            "--view-key", dest="view_key", help="Hex coded private view key"
        )

        parser.add_argument(
            "--account-file",
            dest="account_file",
            help="Account file with watch-only creds",
        )

        parser.add_argument(
            "--watch-wallet", dest="watch_wallet", help="Watch-only wallet files"
        )

        parser.add_argument(
            "--monero-bin", dest="monero_bin", help="Directory with monero binaries"
        )

        parser.add_argument(
            "--rpc-addr", dest="rpc_addr", default=None, help="RPC address of full node"
        )

        parser.add_argument(
            "--rpc-wallet",
            dest="wallet_rpc_addr",
            default=None,
            help="Use running monero-wallet-rpc",
        )

        parser.add_argument(
            "--rpc-wallet-creds",
            dest="wallet_rpc_creds",
            default=None,
            help="Running monero-wallet-rpc credentials",
        )

        parser.add_argument(
            "--sign", dest="sign", default=None, help="Sign the unsigned file"
        )

        parser.add_argument(
            "--debug",
            dest="debug",
            default=False,
            action="store_const",
            const=True,
            help="Debugging output",
        )

        parser.add_argument(
            "--debug-rpc",
            dest="debug_rpc",
            default=False,
            action="store_const",
            const=True,
            help="Prints output of the RPC wallet",
        )

        parser.add_argument(
            "--testnet",
            dest="testnet",
            default=False,
            action="store_const",
            const=True,
            help="Testnet",
        )

        parser.add_argument(
            "--slip0010",
            dest="slip0010",
            default=False,
            action="store_const",
            const=True,
            help="SLIP0010 wallet derivation",
        )

        parser.add_argument(
            "--trezor",
            dest="trezor",
            default=True,
            action="store_const",
            const=True,
            help="Use Trezor connector",
        )

        parser.add_argument(
            "--trezor-path", dest="trezor_path", default=None, help="Trezor path"
        )

        parser.add_argument(
            "--poc",
            dest="poc",
            default=False,
            action="store_const",
            const=True,
            help="Use PoC Trezor emulation",
        )

        parser.add_argument(
            "--patch-client",
            dest="patch_client",
            default=False,
            action="store_const",
            const=True,
            help="Monkey pathing of Trezor client",
        )

        args_src = sys.argv
        self.args = parser.parse_args(args=args_src[1:])

        if self.args.rpc_addr:
            if not re.match(r"^\[?([.0-9a-f:]+)\]?(:[0-9]+)?$", self.args.rpc_addr):
                logger.error("Invalid deamon address: %s" % self.args.rpc_addr)
                return -1

        sys.argv = [args_src[0]]
        res = await self.entry()
        sys.argv = args_src
        return res


async def amain():
    agent = HostAgent()
    res = await agent.main()
    sys.exit(res)


def main():
    loop = asyncio.get_event_loop()
    loop.run_until_complete(amain())
    # loop.run_forever()
    loop.close()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import asyncio
import binascii
import collections
import json
import logging
import os
import random
import re
import signal
import string
import subprocess
import sys
import threading
import time
import traceback
from shlex import quote

import shellescape
from sarge import Capture, Feeder, run

from monero_glue.hwtoken.misc import TrezorError
from monero_glue.xmr import crypto
from monero_glue.xmr.monero import DISPLAY_DECIMAL_POINT

logger = logging.getLogger(__name__)


PRIORITIES = ["default", "unimportant", "normal", "elevated", "priority"]


class TrezorCallError(TrezorError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TrezorNotConfiguredError(TrezorError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TrezorAddressMismatchError(TrezorError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class CliPrompt(object):
    """
    Synchronous CLI confirmation waiter.
    """

    def __init__(self, pre_wait_hook=None, *args, **kwargs):
        self.in_confirmation = False
        self.conf_evt = None
        self.confirmed_result = False
        self.pre_wait_hook = pre_wait_hook

    def confirmation(self, confirmed):
        self.confirmed_result = confirmed
        self.in_confirmation = False
        self.conf_evt.set()

    def _init_wait(self):
        self.in_confirmation = True
        self.confirmed_result = False
        self.conf_evt = threading.Event()

    def wait_confirmation(self):
        """
        Synchronous waiting
        :return:
        """
        self._init_wait()
        if self.pre_wait_hook:
            self.pre_wait_hook()

        try:
            self.conf_evt.wait()
            return self.confirmed_result

        finally:
            self.in_confirmation = False

    async def async_wait_confirmation(self):
        """
        Asynchronous waiting
        Asyncio waiting not implemented.

        :return:
        """
        self._init_wait()
        try:
            # TODO: asyncio wait
            # TODO: eventlet could require another locking mechanism
            self.conf_evt.wait()
            return self.confirmed_result

        finally:
            self.in_confirmation = False

    def __enter__(self):
        self._init_wait()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.in_confirmation = False


def py_raw_input(question=None):
    """
    Python compatibility wrapper for standard raw_input()
    :param question:
    :return:
    """
    return input(question)


def gen_simple_passwd(n):
    """
    Generates simple alphanum passwd.
    TODO: stronger passwd
    :param n:
    :return:
    """
    return "".join(
        random.choice(string.ascii_uppercase + string.digits) for _ in range(n)
    )


def amount_to_uint64(amount):
    return int((10 ** DISPLAY_DECIMAL_POINT) * amount)


class TransferArgs(object):
    def __init__(self):
        self.indices = []
        self.priority = None
        self.mixin = None
        self.address_amounts = []
        self.address = None
        self.amount = None
        self.payment_id = None
        self.amount_threshold = None
        self.key_image = None


def cut_index(parts):
    if not parts[0].startswith('index='):
        return [], parts

    return [int(x.strip()) for x in parts[0][6:].split(',')], parts[1:]


def cut_priority(parts):
    idx = None
    for i, x in enumerate(PRIORITIES):
        if parts[0] == x:
            idx = i
            break

    if idx is not None:
        return idx, parts[1:]
    return None, parts


def cut_mixin(parts):
    try:
        return int(parts[0]), parts[1:]
    except Exception as e:
        return None, parts


def cut_addr(parts):
    return parts[0], parts[1:]


def cut_amount(parts):
    try:
        return float(parts[0]), parts[1:]
    except Exception as e:
        raise ValueError('Format error, Amount not valid') from e


def cut_addr_amount(parts):
    try:
        return parts[0], float(parts[1]), parts[2:]
    except Exception as e:
        return None, None, parts


def cut_key_image(parts):
    ki = parts[0]
    if len(ki) != 64:
        raise ValueError('Format error, Key Image have to be exactly 32B long')
    try:
        return binascii.unhexlify(ki), parts[1:]
    except Exception as e:
        raise ValueError('Format error, Key Image is not properly hex-coded') from e


def cut_payment_id(parts):
    if len(parts) == 0:
        return None, parts
    payment_id = parts[0]
    if len(payment_id) != 16 and len(payment_id) != 64:
        raise ValueError('Payment ID allowed only 8 and 32 B')
    return payment_id, parts[1:]


def parse_transfer_cmd(parts):
    """
    Parses transfer command format
    transfer [index=<N1>[,<N2>,...]] [<priority>] [<ring_size>] <address> <amount> [<payment_id>]
    :param parts:
    :return:
    """
    if len(parts) < 2:
        raise ValueError('Format error, too few arguments')

    parts = list(parts)  # copy
    res = TransferArgs()
    res.indices, parts = cut_index(parts)
    res.priority, parts = cut_priority(parts)
    res.mixin, parts = cut_mixin(parts)
    res.address_amounts = []

    while True:
        addr, amount, parts = cut_addr_amount(parts)
        if addr is None:
            break

        res.address_amounts.append((addr, amount))

    if len(res.address_amounts) == 0:
        raise ValueError('Format error')

    res.payment_id, parts = cut_payment_id(parts)
    return res


def parse_sweep_all(line):
    """
    sweep_all [index=<N1>[,<N2>,...]] [<priority>] [<ring_size>] <address> [<payment_id>]
    :param line:
    :return:
    """
    parts = [x for x in line.split(" ") if len(x.strip()) > 0]
    res = TransferArgs()
    res.indices, parts = cut_index(parts)
    res.priority, parts = cut_priority(parts)
    res.mixin, parts = cut_mixin(parts)
    res.address, parts = cut_addr(parts)
    res.payment_id, parts = cut_payment_id(parts)
    return res


def parse_sweep_below(line):
    """
    sweep_below <amount_threshold> [index=<N1>[,<N2>,...]] [<priority>] [<ring_size>] <address> [<payment_id>]
    :param line:
    :return:
    """
    parts = [x for x in line.split(" ") if len(x.strip()) > 0]
    if len(parts) < 2:
        raise ValueError('Invalid format')

    res = TransferArgs()
    res.amount_threshold, parts = cut_amount(parts)
    res.indices, parts = cut_index(parts)
    res.priority, parts = cut_priority(parts)
    res.mixin, parts = cut_mixin(parts)
    res.address, parts = cut_addr(parts)
    res.payment_id, parts = cut_payment_id(parts)
    return res


def parse_sweep_single(line):
    """
    sweep_single [<priority>] [<ring_size>] <key_image> <address> [<payment_id>]
    :param line:
    :return:
    """
    parts = [x for x in line.split(" ") if len(x.strip()) > 0]
    if len(parts) < 2:
        raise ValueError('Invalid format')

    res = TransferArgs()
    res.priority, parts = cut_priority(parts)
    res.mixin, parts = cut_mixin(parts)
    res.key_image, parts = cut_key_image(parts)
    res.address, parts = cut_addr(parts)
    res.payment_id, parts = cut_payment_id(parts)
    return res


class SargeLogFilter(logging.Filter):
    """Filters out debugging logs generated by sarge - output capture. It is way too verbose for debug"""

    def __init__(self, name="", *args, **kwargs):
        self.namex = name
        logging.Filter.__init__(self, *args, **kwargs)

    def filter(self, record):
        if record.levelno != logging.DEBUG:
            return 1

        try:
            # Parse messages are too verbose, skip.
            if record.name == "sarge.parse":
                return 0

            # Disable output processing message - length of one character.
            msg = record.getMessage()
            if "queued chunk of length 1" in msg:
                return 0

            return 1

        except Exception as e:
            logger.error("Exception in log filtering: %s" % e)

        return 1


def install_sarge_filter():
    """
    Installs Sarge log filter to avoid long 1char debug dumps
    :return:
    """
    for handler in logging.getLogger().handlers:
        handler.addFilter(SargeLogFilter("hnd"))
    logging.getLogger().addFilter(SargeLogFilter("root"))


def sarge_sigint(proc):
    """
    Sends sigint to sarge process
    :return:
    """
    proc.process_ready.wait()
    p = proc.process
    if not p:  # pragma: no cover
        raise ValueError("There is no subprocess")
    p.send_signal(signal.SIGINT)


def escape_shell(inp):
    """
    Shell-escapes input param
    :param inp:
    :return:
    """
    try:
        inp = inp.decode('utf8')
    except:
        pass

    try:
        return shellescape.quote(inp)
    except:
        pass

    quote(inp)


def wallet_enc_key(salt, password):
    """
    Chacha20 password derivation for view key storage
    :param salt:
    :param password:
    :return:
    """
    return crypto.pbkdf2(password, salt, count=2048)



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class TronFrozenSupply(p.MessageType):

    def __init__(
        self,
        frozen_amount: int = None,
        frozen_days: int = None,
    ) -> None:
        self.frozen_amount = frozen_amount
        self.frozen_days = frozen_days

    @classmethod
    def get_fields(cls):
        return {
            1: ('frozen_amount', p.UVarintType, 0),
            2: ('frozen_days', p.UVarintType, 0),
        }



import fileinput
import re
import io
import argparse
import logging
import coloredlogs
import time

logger = logging.getLogger(__name__)
coloredlogs.CHROOT_FILES = []
coloredlogs.install(level=logging.INFO, use_chroot=False)


TICKS_PER_SECOND = 1000000.


def tic2sec(tic):
    return tic/TICKS_PER_SECOND


class LogAnalyzer(object):
    def __init__(self):
        self.max_line_len = 0
        self.time_prev = 0
        self.time_ref = 0
        self.time_ref_r = 0
        self.mem_alloc_prev = 0
        self.mem_alloc_ref = None
        self.mem_alloc_max = 0
        self.serial_device = None

    def process(self, line):
        line = line.strip()
        m = re.match(r'^(\d+)\s([^\s]+?)\s([^\s]+?)\s(.*)$', line)
        if m is None:
            print(line)
            return

        ctime_r = time.time()
        ctime = int(m.group(1))
        c_prev = self.time_prev
        c_prev_alloc = self.mem_alloc_prev
        self.time_prev = ctime

        if '----diagnostic' in line:
            self.time_ref = ctime
            self.time_ref_r = ctime_r
            return

        self.max_line_len = max(self.max_line_len, min(len(line), 140))
        abs_time = tic2sec(ctime - self.time_ref)
        diff_time = tic2sec(ctime - c_prev)

        mem_free = None
        mem_alloc = None

        mmem = re.match(r'.+?F:\s*(\d+)\sA:\s*(\d+)', line)
        if mmem:
            mem_free = int(mmem.group(1))
            mem_alloc = int(mmem.group(2))

        mmem = re.match(r'.+?Free:\s*(\d+)\sAllocated:\s*(\d+)', line)
        if mmem:
            mem_free = int(mmem.group(1))
            mem_alloc = int(mmem.group(2))

        memstr = ''
        if mem_alloc:
            self.mem_alloc_prev = mem_alloc
            self.mem_alloc_max = max(self.mem_alloc_max, mem_alloc)
            if self.mem_alloc_ref is None:
                self.mem_alloc_ref = mem_alloc

            memstr = 'Alloc diff: %5d, refdi: %5d' % (mem_alloc - self.mem_alloc_ref, mem_alloc - c_prev_alloc)

        ldiff = self.max_line_len - len(line)
        print('%s%s |  AbsTime: %7.3f,   Diff %5.3f  | %s' % (line, ' '*ldiff, abs_time, diff_time, memstr))

        if '====' in line or '####' in line:
            abs_r = ''
            if self.serial_device:
                abs_time_r = ctime_r - self.time_ref_r
                abs_r = 'r: %7.2f, ticks p.s.: %7.3f' % (abs_time_r, (ctime - self.time_ref) / float(abs_time_r))
            print(' ++ TOTAL: %7.2f, %s mem max: %s' % (abs_time, abs_r, self.mem_alloc_max - self.mem_alloc_ref))

            self.time_ref = ctime
            self.time_ref_r = ctime_r
            if mem_alloc is not None:
                self.mem_alloc_ref = mem_alloc
                self.mem_alloc_max = mem_alloc

    def read_serial(self, device, brate):
        try:
            import serial  # pip install pyserial
        except ImportError:
            raise ValueError('pip install pyserial')

        while True:
            try:
                ser = serial.Serial(device, brate, timeout=.1)
                logger.info('Connected: %s' % ser)

                sio = io.TextIOWrapper(io.BufferedRWPair(ser, ser))
                while True:
                    line = sio.readline()
                    line = line.strip()
                    if len(line) == 0:
                        continue

                    self.process(line)

            except Exception as e:
                logger.warning('Exc: %s' % e)
                time.sleep(2)


    def read_files(self, files):
        for idx, line in enumerate(fileinput.input(files)):
            anz.process(line)

    def main(self):
        parser = argparse.ArgumentParser(description='Trezor log reader and parser')
        parser.add_argument('--serial', default=None, help='Serial device to read from')
        parser.add_argument('--brate', type=int, default=115200, help='Baud rate')
        parser.add_argument("--retry", dest="retry", default=True, action="store_const", const=True, help="Retry reconnect")
        parser.add_argument('files', metavar='FILE', nargs='*', help='files to read, if empty, stdin is used')
        args = parser.parse_args()

        if args.serial:
            self.serial_device = args.serial
            self.read_serial(args.serial, args.brate)
        else:
            self.read_files(args.files)


if __name__ == '__main__':
    anz = LogAnalyzer()
    anz.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018


MONERO_CURVE = "ed25519-keccak"


async def monero_get_creds(ctx, address_n=None, network_type=None):
    """
    Credentials derivation method.
    In the emulated setup (no trezor environment) this fallbacks to retrieving credentials
    from the context (trezor lite).

    :param ctx:
    :param address_n:
    :param network_type:
    :return:
    """
    return await ctx.monero_get_creds(address_n=address_n, network_type=network_type)


def get_interface(ctx):
    return ctx.get_iface()



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class TronTransferAssetContract(p.MessageType):

    def __init__(
        self,
        asset_name: str = None,
        to_address: bytes = None,
        amount: int = None,
    ) -> None:
        self.asset_name = asset_name
        self.to_address = to_address
        self.amount = amount

    @classmethod
    def get_fields(cls):
        return {
            1: ('asset_name', p.UnicodeType, 0),
            2: ('to_address', p.BytesType, 0),
            3: ('amount', p.UVarintType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class OntologyTxAttribute(p.MessageType):

    def __init__(
        self,
        usage: int = None,
        data: bytes = None,
    ) -> None:
        self.usage = usage
        self.data = data

    @classmethod
    def get_fields(cls):
        return {
            1: ('usage', p.UVarintType, 0),
            2: ('data', p.BytesType, 0),
        }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018
#
# Limbs computation from: https://www.imperialviolet.org/2010/12/04/ecc.html
# Used to implement sub256_modm() - computation of 2xM, 4xM for modular reduction and underflow protection.
#

import math

from monero_glue.xmr import crypto

LIMBS = 9
LIMB_SIZE = 30

MASK = 2 ** LIMB_SIZE - 1


def sumlimb(limbs):
    res = 0
    for i in range(LIMBS):
        res += (2 ** (LIMB_SIZE * i)) * limbs[i]
    return res


def to_limbs(num):
    limbs = []
    for i in range(LIMBS):
        limbs.append(num & (2 ** LIMB_SIZE - 1))
        num >>= LIMB_SIZE
    return limbs


def print_limbs(limbs):
    res = 0
    for i in range(LIMBS):
        res += (2 ** (LIMB_SIZE * i)) * limbs[i]
        print("  %x %s" % (limbs[i], math.log(limbs[i], 2) if limbs[i] > 0 else "-"))
    print("Total: %s, below moduli: %s" % (res, res < crypto.l))


def limbs_consts(var=0, limbs_val=0, limbs=None, value=None):
    print("** Limbs var: %s val: %x" % (var, limbs_val))

    # Compute maximal reduced limb value for < top limb.
    # Top limb has specific (moduli).
    pbase = 2 ** LIMB_SIZE
    if var == 1:
        pbase = 2 ** (LIMB_SIZE + 1)
    if var == 2:
        pbase = 2 ** LIMB_SIZE + 2 ** (LIMB_SIZE + 1)

    if value is not None:
        limbs = to_limbs(value)
    else:
        limbs = [limbs_val] * LIMBS if not limbs else limbs
        value = sumlimb(limbs)

    sums = sumlimb(limbs)
    sumsmod = sums % crypto.l
    print("sum:    0x%x" % sums)
    print("summod: 0x%x" % sumsmod)

    # Top limb
    first = crypto.l >> (LIMB_SIZE * (LIMBS - 1))
    print("Top limb: 0x%x, bit-size: %s\n" % (first, math.log(first, 2)))

    over_idx = LIMBS * LIMB_SIZE
    leftover = pbase >> LIMB_SIZE
    final_leftover = leftover << over_idx

    # First limb is without leftover
    limbs[0] += pbase
    print("Pbase:            0x%x, log: %s" % (pbase, math.log(pbase, 2)))
    print("Leftover:         0x%x" % leftover)

    for i in range(1, LIMBS):
        if i < LIMBS - 1:
            limbs[i] += pbase - leftover
        else:
            # Trick to reduce number of modular reductions
            limbs[i] += first // 2 - leftover
            final_leftover = (first // 2) << (LIMB_SIZE * (LIMBS - 1))
            # limbs[i] += first - leftover
            # final_leftover = (first) << (LIMB_SIZE*(LIMBS-1))

    print("Final_leftover:   0x%x" % final_leftover)

    left_moduled = final_leftover - crypto.l
    left_moduled = -left_moduled % crypto.l
    print("Test:             %s" % bool(final_leftover > crypto.l))
    print("Moduli:           0x%x" % crypto.l)
    print("Moduled leftover: 0x%x" % left_moduled)

    for i in range(LIMBS):
        limbs[i] += (left_moduled >> (i * LIMB_SIZE)) & ((2 ** LIMB_SIZE) - 1)

    # limbs print
    print("\nLimbs: ")
    print("[" + (", ".join(["0x%x" % x for x in limbs])) + "]\n")

    print("Limbs sizes: ")
    res = 0
    for i in range(LIMBS):
        res += (2 ** (LIMB_SIZE * i)) * limbs[i]
        print("  %x %s" % (limbs[i], math.log(limbs[i], 2)))

    print("Final value:     0x%x" % res)
    print("Final value mod: 0x%x" % (res % crypto.l))
    print("Final correct:   %s" % (res % crypto.l - sumsmod == 0))

    # collapse with carry
    print("\nLimbs with carry sum: ")
    c = 0
    res = 0
    for i in range(LIMBS):
        c += limbs[i]
        res += (2 ** (LIMB_SIZE * i)) * (c & MASK)
        c >>= LIMB_SIZE

        print("  %x %s c: %x" % (limbs[i], math.log(limbs[i], 2), c))

    print("")
    print("Final value:     0x%x" % res)
    print("Final value mod: 0x%x" % (res % crypto.l))
    print(
        "Final value mul: 0x%x" % (res // crypto.l)
    )  # number of modular reductions needed
    print("CORRECT: %s" % ((res % crypto.l) == value))
    print("-" * 80)


limbs_consts(0, value=0)
limbs_consts(1, value=0)

print("=" * 80)
# limbs_consts(0, value=crypto.l-1)
# limbs_consts(1, value=crypto.l-1)

lmbs = [
    0x5cf5d3ed,
    0x60498c68,
    0x6f79cd64,
    0x77be77a7,
    0x40000013,
    0x3fffffff,
    0x3fffffff,
    0x3fffffff,
    0xfff,
]
limbs = [0x0fffffff, 0x0fffffff, 0x0fffffff, 0x0fffffff, 4, 0, 0, 0, 0x1000]

print_limbs(limbs)
limbs_consts(0, limbs=limbs)

print_limbs(to_limbs(crypto.l - 1))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

from monero_glue import messages, protobuf
from monero_glue.messages import *


def get_message_type(msg=None, cls=None):
    if msg is not None:
        cls = msg.__class__

    type_id = getattr(MessageType, cls.__name__)
    if type_id is None:
        raise ValueError("Could not find message ID for: %s" % msg)
    return type_id


def get_message_from_type(type_id):
    all_msg = [x for x in dir(MessageType) if not x.startswith("_")]
    for m in all_msg:
        cur_id = getattr(MessageType, m)
        if cur_id == type_id:
            return getattr(messages, m)
    raise ValueError("Could not find message with type: %s" % type_id)


def bytes_fix(msg, pkg=protobuf):
    """
    Converts all elements to bytes if specified by protobuf scheme
    :param msg:
    :param pkg:
    :return:
    """
    if isinstance(msg, pkg.MessageType):
        fields = msg.get_fields()
        for tag in fields:
            field = fields[tag]
            field_name = field[0]
            if hasattr(msg, field_name):
                cval = getattr(msg, field_name, None)
                if cval is not None and field[1] == pkg.BytesType:
                    if field[2] & pkg.FLAG_REPEATED:
                        cval = [bytes(x) for x in cval]
                    else:
                        cval = bytes(cval)
                setattr(msg, field_name, bytes_fix(cval, pkg))
        return msg

    elif isinstance(msg, bytearray):
        return bytes(msg)
    elif isinstance(msg, list):
        return [bytes_fix(x, pkg) for x in msg]
    else:
        return msg


class MessageConverter(object):
    """
    Converts protobuf messages between base packages.
    Converts our protobufs <-> trezorlib
    Required for protobuf encoders
    """

    def __init__(self, fix_bytes=False):
        self.imported = False
        self.tlib_msgs = None
        self.tlib_p = None
        self.fix_bytes = fix_bytes

    def _init(self):
        if self.imported:
            return

        from trezorlib import messages as tlib_msgs
        from trezorlib import protobuf as tlib_protobuf

        self.tlib_msgs = tlib_msgs
        self.tlib_p = tlib_protobuf
        self.imported = True

    def _transform(self, msg, src_proto, dst_package):
        if isinstance(msg, src_proto.MessageType):
            msg_name = msg.__class__.__name__
            tlib_msg = getattr(dst_package, msg_name)

            nmsg = tlib_msg()
            fields = msg.get_fields()
            for tag in fields:
                field = fields[tag]
                field_name = field[0]
                if hasattr(msg, field_name):
                    cval = getattr(msg, field_name, None)
                    setattr(nmsg, field_name, self.to_trezorlib(cval))
            return nmsg

        elif isinstance(msg, list):
            return [self.to_trezorlib(x) for x in msg]
        elif isinstance(msg, dict):
            return {v: self.to_trezorlib(x) for v, x in msg.items()}
        else:
            return msg

    def to_phlib(self, msg):
        self._init()
        if self.fix_bytes:
            msg = bytes_fix(msg, self.tlib_p)
        return self._transform(msg, self.tlib_p, messages)

    def to_trezorlib(self, msg):
        self._init()
        if self.fix_bytes:
            msg = bytes_fix(msg, protobuf)
        res = self._transform(msg, protobuf, self.tlib_msgs)
        return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import sys


def memcpy(dst, dst_off, src, src_off, len):
    for i in range(len):
        dst[dst_off + i] = src[src_off + i]
    return dst


def unimport_begin():
    return set(sys.modules)


def unimport_end(mods):
    pass



import random
import re
import io
import os
import argparse
import logging
import coloredlogs
import time
import asyncio
import sys
import binascii

from monero_serialize.helpers import ArchiveException
from monero_serialize.xmrtypes import TxExtraPubKey, TxExtraNonce, TxExtraAdditionalPubKeys, TxDestinationEntry

from monero_glue.hwtoken import misc
from monero_glue.xmr import wallet, crypto, monero
from monero_glue.xmr.monero import XmrNoSuchAddressException, generate_key_image_helper_precomp
from monero_glue.xmr.sub import addr
from monero_glue.xmr.sub.creds import AccountCreds

logger = logging.getLogger(__name__)
coloredlogs.CHROOT_FILES = []
coloredlogs.install(level=logging.DEBUG, use_chroot=False)


def generate_key_image_ex(
    creds,
    subaddresses,
    out_key,
    tx_public_key,
    additional_tx_public_keys,
    real_output_index,
):
    recv_derivation = crypto.generate_key_derivation(tx_public_key, creds.view_key_private)

    additional_recv_derivations = []
    for add_pub_key in additional_tx_public_keys:
        additional_recv_derivations.append(
            crypto.generate_key_derivation(add_pub_key, creds.view_key_private)
        )

    subaddr_recv_info = monero.is_out_to_acc_precomp(
        subaddresses,
        out_key,
        recv_derivation,
        additional_recv_derivations,
        real_output_index,
    )
    if subaddr_recv_info is None:
        raise XmrNoSuchAddressException()

    xi, ki = generate_key_image_helper_precomp(
        creds, out_key, subaddr_recv_info[1], real_output_index, subaddr_recv_info[0]
    )
    return xi, ki, recv_derivation, subaddr_recv_info


class TestGen(object):
    def __init__(self):
        self.args = None
        self.dest_keys = None  # type: AccountCreds
        self.cur_keys = None  # type: AccountCreds
        self.cur_subs = {}
        self.dest_subs = {}

        self.add_additionals = False
        self.dest_sub_major = 0

    async def process(self, idx, inp):
        is_stdin = False
        if inp == b"-":
            file = getattr(sys.stdin, "buffer", sys.stdin)
            is_stdin = True
        else:
            file = open(inp, "rb")

        try:
            logger.debug("Processing: %s" % inp)
            data = file.read()
            unsigned_tx = await wallet.load_unsigned_tx(self.cur_keys.view_key_private, data)

            if self.args.desc:
                await self.describe(inp, unsigned_tx, self.cur_keys, self.cur_subs)

            if self.dest_keys is None:
                return

            new_unsigned = await self.rekey_unsigned(unsigned_tx)

            new_bin = await wallet.dump_unsigned_tx(self.dest_keys.view_key_private, new_unsigned)

            if self.args.output:
                nbase = 'output.bin' if is_stdin else os.path.basename(inp)
                nbase_fname, nbase_ext = os.path.splitext(nbase)
                if self.args.suffix:
                    nbase_fname += self.args.suffix

                nbase = nbase_fname + nbase_ext
                ndest = os.path.join(self.args.output, nbase)

                if self.args.desc:
                    await self.describe(ndest, new_unsigned, self.dest_keys, self.dest_subs)

                with open(ndest, 'wb') as fh:
                    fh.write(new_bin)
                logger.info('Result written to: %s' % ndest)

            else:
                sys.stdout.buffer.write(new_bin)

        except ArchiveException as ae:
            logger.error(ae)
            logger.error(ae.tracker)

        finally:
            if not is_stdin:
                file.close()

    async def analyze_input(self, keys, subs, inp):
        subs = subs if subs else {}
        real_out_key = inp.outputs[inp.real_output][1]
        out_key = crypto.decodepoint(real_out_key.dest)
        tx_key = crypto.decodepoint(inp.real_out_tx_key)
        additional_keys = [
            crypto.decodepoint(x) for x in inp.real_out_additional_tx_keys
        ]

        res = generate_key_image_ex(keys, subs, out_key,
            tx_key,
            additional_keys,
            inp.real_output_in_tx_index,)

        xi, ki, recv_derivation, subaddr_recv_info = res
        sub_idx = subaddr_recv_info[0]
        return sub_idx

    async def reformat_out(self, out):
        res = misc.StdObj(amount=out.amount,
                          addr=addr.build_address(out.addr.m_spend_public_key,
                                                  out.addr.m_view_public_key),
                          is_subaddress=out.is_subaddress)
        return res

    async def reformat_outs(self, dsts):
        out_txs2 = []
        for o in dsts:
            out_txs2.append(await self.reformat_out(o))
        return out_txs2

    async def describe(self, inp, unsigned_txs, keys, key_subs):
        print('\nInp: %s, #txs: %s' % (inp, len(unsigned_txs.txes)))
        for txid, tx in enumerate(unsigned_txs.txes):
            srcs = tx.sources
            dsts = tx.splitted_dsts
            extra = tx.extra
            change = tx.change_dts
            account = tx.subaddr_account
            subs = tx.subaddr_indices
            amnt_in = sum([x.amount for x in srcs])
            amnt_out = sum([x.amount for x in dsts])
            fee = amnt_in - amnt_out
            n_inp_additional = sum([1 for x in srcs if len(x.real_out_additional_tx_keys) > 0])

            change_addr = addr.build_address(change.addr.m_spend_public_key, change.addr.m_view_public_key) if change else None
            out_txs2 = await self.reformat_outs(dsts)

            num_stdaddresses, num_subaddresses, single_dest_subaddress = addr.classify_subaddresses(out_txs2, change_addr)

            print('  tx: %s, #inp: %2d, #inp_add: %2d, #out: %2d, acc: %s, subs: %s, '
                  'xmr_in: %10.6f, xmr_out: %10.6f, fee: %10.6f, change: %10.6f, out_clean: %10.6f'
                  % (txid, len(srcs), n_inp_additional, len(dsts),
                     account, subs,
                     wallet.conv_disp_amount(amnt_in),
                     wallet.conv_disp_amount(amnt_out),
                     wallet.conv_disp_amount(fee),
                     wallet.conv_disp_amount(change.amount) if change else 0,
                     wallet.conv_disp_amount((amnt_out - change.amount) if change else amnt_out),
                     ))
            print('  Out: num_std: %2d, num_sub: %2d, single_dest_sub: %s'
                  % (num_stdaddresses, num_subaddresses, 1 if single_dest_subaddress else 0))

            accounts = set()
            subs = set()
            for inp in srcs:
                res = await self.analyze_input(keys, key_subs, inp)
                accounts.add(res[0])
                if res != (0, 0):
                    subs.add(res)

            print('  Ins: accounts: %s, subs: %s' % (accounts, len(subs)))

            extras = await monero.parse_extra_fields(extra)
            extras_val = []
            for c in extras:
                if isinstance(c, TxExtraPubKey):
                    extras_val.append('TxKey')
                elif isinstance(c, TxExtraNonce):
                    extras_val.append('Nonce: %s' % binascii.hexlify(c.nonce).decode('ascii'))
                elif isinstance(c, TxExtraAdditionalPubKeys):
                    extras_val.append('AdditionalTxKeys: %s' % len(c.data))
                else:
                    extras_val.append(str(c))
            print('  Extras: %s' % ', '.join(extras_val))

    async def primary_change_address(self, key, account):
        D, C = monero.generate_sub_address_keys(
            key.view_key_private,
            crypto.scalarmult_base(key.spend_key_private),
            account,
            0,
        )
        return misc.StdObj(
            view_public_key=crypto.encodepoint(C),
            spend_public_key=crypto.encodepoint(D),
        )

    async def adjust_change(self, tx):
        logger.debug('Change addr adjust')
        out_txs2 = await self.reformat_outs(tx.splitted_dsts)
        change_dts = await self.reformat_out(tx.change_dts) if tx.change_dts else None
        change_idx = addr.get_change_addr_idx(out_txs2, change_dts)
        if change_idx is None:
            return

        change_addr = await self.primary_change_address(self.dest_keys, self.dest_sub_major)
        tx.change_dts.addr.m_spend_public_key = change_addr.spend_public_key
        tx.change_dts.addr.m_view_public_key = change_addr.view_public_key

        tx.splitted_dsts[change_idx].addr.m_spend_public_key = change_addr.spend_public_key
        tx.splitted_dsts[change_idx].addr.m_view_public_key = change_addr.view_public_key

    async def amplify_outs(self, tx):
        lst = tx.splitted_dsts[-1]
        orig_amount = lst.amount
        partial = orig_amount // (self.args.outs + 1)
        amnt_sum = partial
        lst.amount = partial

        for i in range(1, self.args.outs + 1):
            is_lst = i >= self.args.outs
            new_amnt = orig_amount - amnt_sum if is_lst else partial
            amnt_sum += partial
            new_tx = TxDestinationEntry(amount=new_amnt, is_subaddress=lst.is_subaddress, addr=lst.addr)
            tx.splitted_dsts.append(new_tx)

    async def rekey_unsigned(self, unsigned_txs):
        for tx in unsigned_txs.txes:
            tx.use_bulletproofs = False
            logger.debug(
                "Transaction with %s inputs, %s outputs, %s mix ring"
                % (len(tx.sources), len(tx.splitted_dsts), len(tx.sources[0].outputs))
            )

            for inp in tx.sources:
                self.rekey_input(inp, self.cur_keys, self.cur_subs, self.dest_keys, self.dest_subs)

            if tx.change_dts and tx.subaddr_account != self.dest_sub_major:
                await self.adjust_change(tx)

            tx.subaddr_account = self.dest_sub_major
            tx.subaddr_indices = self.args.minors

            if self.args.outs is not None:
                await self.amplify_outs(tx)

        return unsigned_txs

    def rekey_input(self, inp, keys, subs=None, new_keys=None, new_subs=None):
        subs = subs if subs else {}
        real_out_key = inp.outputs[inp.real_output][1]
        out_key = crypto.decodepoint(real_out_key.dest)
        tx_key = crypto.decodepoint(inp.real_out_tx_key)
        additional_keys = [
            crypto.decodepoint(x) for x in inp.real_out_additional_tx_keys
        ]

        logger.debug('Current out key: %s' % binascii.hexlify(real_out_key.dest))
        secs = monero.generate_key_image_helper(
            keys,
            subs,
            out_key,
            tx_key,
            additional_keys,
            inp.real_output_in_tx_index,
        )
        xi, ki, di = secs

        need_additional = additional_keys is not None and len(additional_keys) > 0
        is_dst_sub = self.dest_sub_major != 0 and (self.args.minors[0] != 0 or len(self.args.minors) > 1)
        logger.debug('Is dst sub: %s, need additional: %s' % (is_dst_sub, need_additional))

        if is_dst_sub and self.add_additionals:
            need_additional = True

        if is_dst_sub:
            rand_minor = random.choice(self.args.minors)
            m = monero.get_subaddress_secret_key(new_keys.view_key_private, major=self.dest_sub_major, minor=rand_minor)
            M = crypto.scalarmult_base(m)
            d = crypto.sc_add(m, new_keys.spend_key_private)
            D = crypto.point_add(new_keys.spend_key_public, M)
            C = crypto.scalarmult(D, new_keys.view_key_private)

        if not need_additional and not is_dst_sub:
            # real_out_key.dst = Hs(R*new_a || idx)G + newB
            r = crypto.random_scalar()
            tx_key = crypto.scalarmult_base(r)
            new_deriv = crypto.generate_key_derivation(new_keys.view_key_public, r)
            new_out_pr = crypto.derive_secret_key(new_deriv, inp.real_output_in_tx_index, new_keys.spend_key_private)
            new_out = crypto.scalarmult_base(new_out_pr)
            real_out_key.dest = crypto.encodepoint(new_out)

        elif not need_additional and is_dst_sub:
            # real_out_key.dst = Hs(r*C || idx)G + newB, R=rD
            r = crypto.random_scalar()
            tx_key = crypto.scalarmult(D, r)
            new_deriv = crypto.generate_key_derivation(C, r)
            new_out_pr = crypto.derive_secret_key(new_deriv, inp.real_output_in_tx_index, d)
            new_out = crypto.scalarmult_base(new_out_pr)
            real_out_key.dest = crypto.encodepoint(new_out)

        else:
            r = crypto.random_scalar()
            tx_key = crypto.scalarmult_base(r)

            gen_additionals = min(2, inp.real_output_in_tx_index + 1)
            if additional_keys is None or len(additional_keys) < gen_additionals:
                additional_keys = [crypto.scalarmult_base(crypto.random_scalar()) for _ in range(gen_additionals)]

            ri = crypto.random_scalar()
            if is_dst_sub:
                add_tx = crypto.scalarmult(D, ri)
                new_deriv = crypto.generate_key_derivation(C, ri)
                new_out_pr = crypto.derive_secret_key(new_deriv, inp.real_output_in_tx_index, d)
                new_out = crypto.scalarmult_base(new_out_pr)
                if not crypto.point_eq(new_out, crypto.derive_public_key(new_deriv, inp.real_output_in_tx_index, D)):
                    raise ValueError('Invalid txout computation')

            else:
                add_tx = crypto.scalarmult_base(ri)
                new_deriv = crypto.generate_key_derivation(new_keys.view_key_public, r)
                new_out_pr = crypto.derive_secret_key(new_deriv, inp.real_output_in_tx_index,
                                                      new_keys.spend_key_private)
                new_out = crypto.scalarmult_base(new_out_pr)

            additional_keys[inp.real_output_in_tx_index] = add_tx
            real_out_key.dest = crypto.encodepoint(new_out)

        inp.real_out_tx_key = crypto.encodepoint(tx_key)
        inp.real_out_additional_tx_keys = [crypto.encodepoint(x) for x in additional_keys]

        logger.debug('New pub: %s' % binascii.hexlify(real_out_key.dest))

        # Self-check
        secs = monero.generate_key_image_helper(
            new_keys,
            new_subs,
            crypto.decodepoint(real_out_key.dest),
            tx_key,
            additional_keys,
            inp.real_output_in_tx_index,
        )
        return inp

    def precompute_subaddr(self, keys, subs, major_cnt=10, minor_cnt=200):
        if keys is None:
            return None
        for i in range(major_cnt):
            monero.compute_subaddresses(keys, i, list(range(0, minor_cnt)), subs)

    def parse_keys(self, keyvar, ktype="current"):
        if keyvar is None:
            return None
        if len(keyvar) != 2 * 2 * 32:
            raise ValueError(
                "Keys for the %s destination length invalid. Format: SpendKey || ViewKey hexcoded"
                % ktype
            )
        keys_bin = binascii.unhexlify(keyvar)
        spend_key = crypto.decodeint(keys_bin[:32])
        view_key = crypto.decodeint(keys_bin[32:64])

        return AccountCreds.new_wallet(view_key, spend_key)

    async def work(self):
        self.cur_keys = self.parse_keys(self.args.current)
        self.dest_keys = self.parse_keys(self.args.dest, "destination")
        self.dest_sub_major = self.args.major
        self.add_additionals = self.args.add_extra
        self.precompute_subaddr(self.cur_keys, self.cur_subs)
        self.precompute_subaddr(self.dest_keys, self.dest_subs)

        files = self.args.files if self.args.files and len(self.args.files) else [b"-"]
        for idx, line in enumerate(files):
            await self.process(idx, line)

    async def main(self):
        parser = argparse.ArgumentParser(
            description="Test generator - changing unsigned transactions"
        )
        parser.add_argument(
            "--dest",
            default=None,
            required=False,
            help="Destination address to transform to",
        )

        parser.add_argument(
            "--current", default=None, required=True, help="Current destination address"
        )

        parser.add_argument(
            "--output",
            default=None,
            help="Output directory to write results to",
        )

        parser.add_argument(
            "--major",
            default=0,
            type=int,
            help="Destination major address index",
        )

        parser.add_argument(
            "--minors",
            default=[0],
            type=int,
            nargs="*",
            help="Destination minor address index",
        )

        parser.add_argument(
            "--desc",
            dest="desc",
            default=False,
            action="store_const",
            const=True,
            help="Generate test file descriptors",
        )

        parser.add_argument(
            "--suffix",
            dest="suffix",
            default="",
            help="Generated tsx files suffix",
        )

        parser.add_argument(
            "--outs",
            dest="outs",
            default=None,
            type=int,
            help="Amplify the last output to outs - increases tx outputs",
        )

        parser.add_argument('--add-extra', dest='add_extra', action='store_const', const=True, default=False,
                            help='Adds additional tx keys')

        parser.add_argument(
            "files",
            metavar="FILE",
            nargs="*",
            help="files to read, if empty, stdin is used",
        )
        self.args = parser.parse_args()
        return await self.work()


async def amain():
    tgen = TestGen()
    res = await tgen.main()
    sys.exit(res)


def main():
    loop = asyncio.get_event_loop()
    loop.run_until_complete(amain())
    loop.close()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import argparse
import binascii

from monero_glue.xmr import crypto, monero


def main():
    """
    Entry point
    :return:
    """
    parser = argparse.ArgumentParser(
        description="Generate non-deterministic wallet credentials"
    )

    parser.add_argument(
        "--testnet",
        dest="testnet",
        default=False,
        action="store_const",
        const=True,
        help="Testnet",
    )

    parser.add_argument(
        "--stagenet",
        dest="stagenet",
        default=False,
        action="store_const",
        const=True,
        help="Testnet",
    )

    args = parser.parse_args()

    network_type = monero.NetworkTypes.MAINNET
    if args.testnet:
        network_type = monero.NetworkTypes.TESTNET
    elif args.stagenet:
        network_type = monero.NetworkTypes.STAGENET

    priv_view = crypto.random_scalar()
    priv_spend = crypto.random_scalar()
    w = monero.AccountCreds.new_wallet(priv_view, priv_spend, network_type=network_type)

    print("Address: %s" % w.address.decode("utf8"))
    print(
        "Private view key:  %s"
        % binascii.hexlify(crypto.encodeint(priv_view)).decode("utf8")
    )
    print(
        "Private spend key: %s"
        % binascii.hexlify(crypto.encodeint(priv_spend)).decode("utf8")
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import traceback

from monero_glue.hwtoken import iface, misc
from monero_glue.messages import (
    DebugMoneroDiagAck,
    Failure,
    FailureType,
    MoneroGetWatchKey,
    MoneroWatchKey,
    MessageType,
)
from monero_glue.protocol.key_image_sync import KeyImageSync
from monero_glue.protocol.tsx_sign import TsxSigner
from monero_glue.protocol_base.error import exc2str
from monero_glue.xmr import crypto, monero


class TokenLite(object):
    """
    Main Trezor object.
    Provides interface to the host, packages messages.
    """

    def __init__(self):
        self.tsx_ctr = 0
        self.err_ctr = 0
        self.tsx_obj = None  # type: TsxSigner
        self.tsx_state = None
        self.ki_sync = None  # type: KeyImageSync
        self.creds = None  # type: monero.AccountCreds
        self.iface = iface.TokenInterface()
        self.debug = True

    async def ki_exc_handler(self, e):
        """
        Handles the exception thrown in the Trezor processing.

        :param e:
        :return:
        """
        if self.debug:
            traceback.print_exc()

        self.err_ctr += 1
        self.ki_sync = None  # clear transaction object
        await self.iface.transaction_error(e)

    async def tsx_exc_handler(self, e):
        """
        Handles the exception thrown in the Trezor processing. Clears transaction state.
        We could use decorator/wrapper for message calls but not sure how uPython handles them
        so now are entry points wrapped in try-catch.

        :param e:
        :return:
        """
        if self.debug:
            traceback.print_exc()

        self.err_ctr += 1
        self.tsx_obj = None  # clear transaction object
        await self.iface.transaction_error(e)

    async def monero_get_creds(self, address_n=None, network_type=None):
        """

        :param network_type:
        :return:
        """
        return self.creds

    def get_iface(self):
        """

        :return:
        """
        return self.iface

    async def test_pb_msg(self, msg):
        """
        Test message serialization
        :return:
        """
        if not __debug__:
            return

        pb = await misc.dump_pb_msg(msg)
        await misc.parse_pb_msg(pb, msg.__class__)

    async def call(self, msg, recode=True):
        return Failure(code=FailureType.FirmwareError, message="unsupported")

    async def call_in_session(self, msg, recode=True):
        return Failure(code=FailureType.FirmwareError, message="unsupported")

    async def ping(self, message=None, **kwargs):
        return DebugMoneroDiagAck()

    async def get_view_key(self, msg: MoneroGetWatchKey):
        if msg.network_type != self.creds.network_type:
            return Failure(message="InvalidNetworkType")

        return MoneroWatchKey(
            watch_key=crypto.encodepoint(self.creds.view_key_private),
            address=self.creds.address,
        )

    async def tsx_sign(self, msg):
        try:
            await self.test_pb_msg(msg)

            signer = TsxSigner()
            await signer.wake_up(self, self.tsx_state, msg, iface=self.iface)
            self.tsx_state = None

            res = await signer.sign(msg)
            if not await signer.should_purge():
                self.tsx_state = await signer.state_save()

            await self.test_pb_msg(res)
            return res

        except Exception as e:
            await self.tsx_exc_handler(e)
            self.tsx_obj = None
            return Failure(message=exc2str(e))

    async def key_image_sync(self, msg):
        try:  # TODO
            if msg.MESSAGE_WIRE_TYPE == MessageType.MoneroKeyImageExportInitRequest:
                self.ki_sync = KeyImageSync(
                    ctx=self, iface=self.iface, creds=self.creds
                )
                return await self.ki_sync.init(self, msg)

            elif msg.MESSAGE_WIRE_TYPE == MessageType.MoneroKeyImageSyncStepRequest:
                return await self.ki_sync.sync(self, msg)

            elif msg.MESSAGE_WIRE_TYPE == MessageType.MoneroKeyImageSyncFinalRequest:
                res = await self.ki_sync.final(self, msg)
                self.ki_sync = None
                return res

            else:
                raise ValueError("Unknown error")

        except Exception as e:
            await self.ki_exc_handler(e)
            self.ki_sync = None
            return Failure(message=exc2str(e))



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .TronFrozenSupply import TronFrozenSupply

if __debug__:
    try:
        from typing import List
    except ImportError:
        List = None  # type: ignore


class TronAssetIssueContract(p.MessageType):

    def __init__(
        self,
        name: str = None,
        abbr: str = None,
        total_supply: int = None,
        frozen_supply: List[TronFrozenSupply] = None,
        trx_num: int = None,
        num: int = None,
        start_time: int = None,
        end_time: int = None,
        description: str = None,
        url: str = None,
    ) -> None:
        self.name = name
        self.abbr = abbr
        self.total_supply = total_supply
        self.frozen_supply = frozen_supply if frozen_supply is not None else []
        self.trx_num = trx_num
        self.num = num
        self.start_time = start_time
        self.end_time = end_time
        self.description = description
        self.url = url

    @classmethod
    def get_fields(cls):
        return {
            2: ('name', p.UnicodeType, 0),
            3: ('abbr', p.UnicodeType, 0),
            4: ('total_supply', p.UVarintType, 0),
            5: ('frozen_supply', TronFrozenSupply, p.FLAG_REPEATED),
            6: ('trx_num', p.UVarintType, 0),
            7: ('num', p.UVarintType, 0),
            8: ('start_time', p.UVarintType, 0),
            9: ('end_time', p.UVarintType, 0),
            10: ('description', p.UnicodeType, 0),
            11: ('url', p.UnicodeType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class TronProposalParameters(p.MessageType):

    def __init__(
        self,
        key: int = None,
        value: int = None,
    ) -> None:
        self.key = key
        self.value = value

    @classmethod
    def get_fields(cls):
        return {
            1: ('key', p.UVarintType, 0),
            2: ('value', p.UVarintType, 0),
        }



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p

from .MoneroTransactionDestinationEntry import MoneroTransactionDestinationEntry
from .MoneroTransactionRsigData import MoneroTransactionRsigData


class MoneroTransactionSetOutputRequest(p.MessageType):
    MESSAGE_WIRE_TYPE = 511

    def __init__(
        self,
        dst_entr: MoneroTransactionDestinationEntry = None,
        dst_entr_hmac: bytes = None,
        rsig_data: MoneroTransactionRsigData = None,
    ) -> None:
        self.dst_entr = dst_entr
        self.dst_entr_hmac = dst_entr_hmac
        self.rsig_data = rsig_data

    @classmethod
    def get_fields(cls):
        return {
            1: ('dst_entr', MoneroTransactionDestinationEntry, 0),
            2: ('dst_entr_hmac', p.BytesType, 0),
            3: ('rsig_data', MoneroTransactionRsigData, 0),
        }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018


from monero_glue.messages import (
    Failure,
    FailureType,
    MoneroExportedKeyImage,
    MoneroKeyImageExportInitAck,
    MoneroKeyImageExportInitRequest,
    MoneroKeyImageSyncFinalAck,
    MoneroKeyImageSyncStepAck,
    MoneroKeyImageSyncStepRequest,
)
from monero_glue.trezor import wrapper as twrap
from monero_glue.xmr import common, crypto, key_image, monero
from monero_glue.xmr.enc import chacha_poly


class KeyImageSync(object):
    def __init__(self, ctx=None, iface=None, creds=None):
        self.ctx = ctx
        self.iface = iface  # type: trezor_iface.TrezorInterface
        self.creds = creds  # type: monero.AccountCreds

        self.num = 0
        self.c_idx = -1
        self.hash = None
        self.blocked = None
        self.enc_key = None
        self.subaddresses = {}
        self.hasher = common.HashWrapper(crypto.get_keccak())

    async def derive_creds(self, msg: MoneroKeyImageExportInitRequest):
        self.creds = await twrap.monero_get_creds(
            self.ctx, msg.address_n or (), msg.network_type
        )

    async def init(self, ctx, msg: MoneroKeyImageExportInitRequest):
        self.ctx = ctx
        await self.derive_creds(msg)

        confirmation = await self.iface.confirm_ki_sync(msg, ctx=ctx)
        if not confirmation:
            return Failure(code=FailureType.ActionCancelled, message="rejected")

        self.num = msg.num
        self.hash = msg.hash
        self.enc_key = crypto.random_bytes(32)

        # Sub address precomputation
        if msg.subs and len(msg.subs) > 0:
            for sub in msg.subs:
                monero.compute_subaddresses(
                    self.creds, sub.account, sub.minor_indices, self.subaddresses
                )
        return MoneroKeyImageExportInitAck()

    async def sync(self, ctx, tds: MoneroKeyImageSyncStepRequest):
        self.ctx = ctx
        if self.blocked:
            raise ValueError("Blocked")
        if len(tds.tdis) == 0:
            raise ValueError("Empty")

        resp = []
        buff = bytearray(32 * 3)
        buff_mv = memoryview(buff)

        for td in tds.tdis:
            self.c_idx += 1
            if self.c_idx >= self.num:
                raise ValueError("Too many outputs")

            hash = key_image.compute_hash(td)
            self.hasher.update(hash)

            ki, sig = await key_image.export_key_image(
                self.creds, self.subaddresses, td
            )

            crypto.encodepoint_into(buff_mv[0:32], ki)
            crypto.encodeint_into(buff_mv[32:64], sig[0][0])
            crypto.encodeint_into(buff_mv[64:], sig[0][1])

            nonce, ciph, tag = chacha_poly.encrypt(self.enc_key, buff)
            eki = MoneroExportedKeyImage(iv=nonce, tag=tag, blob=ciph)
            resp.append(eki)
        return MoneroKeyImageSyncStepAck(kis=resp)

    async def final(self, ctx, msg=None):
        self.ctx = ctx
        if self.blocked:
            raise ValueError("Blocked")

        if self.c_idx + 1 != self.num:
            await self.iface.ki_error("Invalid number of outputs", ctx=self.ctx)
            raise ValueError("Invalid number of outputs")

        final_hash = self.hasher.digest()
        if final_hash != self.hash:
            await self.iface.ki_error("Invalid hash", ctx=self.ctx)
            raise ValueError("Invalid hash")

        return MoneroKeyImageSyncFinalAck(enc_key=self.enc_key)
